from __future__ import annotations
from typing import Any, Optional, Union

import collections
import copy
import numpy as np


__all__ = ["Dimension", "Quantity"]


class Dimension(collections.abc.MutableMapping):
    def __init__(
        self,
        L: Optional[int] = 0,
        T: Optional[int] = 0,
        M: Optional[int] = 0,
        I: Optional[int] = 0,
        K: Optional[int] = 0,
        N: Optional[int] = 0,
        J: Optional[int] = 0,
    ) -> None:

        self._d = {
            "L": L,
            "T": T,
            "M": M,
            "I": I,
            "K": K,
            "N": N,
            "J": J,
        }

    def __iter__(self):
        # FIXME: how to type annotate this?
        return iter(self._d)

    def __getitem__(self, k: str) -> int:
        return self._d.__getitem__(k)

    def __setitem__(self, k: str, v: int) -> None:
        self._d.__setitem__(k, v)

    def __delitem__(self, k: str) -> None:
        del self[k]

    def __len__(self) -> int:
        return 7

    # MULTIPLICATION

    def __mul__(self, other: Dimension) -> Dimension:
        return Dimension(**{k: self._d[k] + other._d[k] for k in self._d})

    def __rmul__(self, other: Dimension) -> Dimension:
        return Dimension(**{k: self._d[k] + other._d[k] for k in self._d})

    def __imul__(self, other: Dimension) -> Dimension:
        for k, v in other._d.items():
            self._d[k] += v

        return self

    # DIVISION

    def __truediv__(self, other: Dimension) -> Dimension:
        return Dimension(**{k: self._d[k] - other._d[k] for k in self._d})

    def __rtruediv__(self, other: Dimension) -> Dimension:
        return Dimension(**{k: other._d[k] - self._d[k] for k in self._d})

    def __itruediv__(self, other: Dimension) -> Dimension:
        for k, v in other._d.items():
            self._d[k] -= v

        return self

    # EXPONENTIATION

    def __pow__(self, other: Dimension) -> Dimension:
        return Dimension(**{k: v * other for k, v in self._d.items()})

    def __ipow__(self, other: Dimension) -> Dimension:
        for k in self._d:
            self._d[k] *= other

        return self

    # COMPARISON

    def __eq__(self, other: Dimension) -> bool:
        return self._d == other._d

    def __hash__(self) -> int:
        return hash(tuple(sorted(self._d.items())))

    def __str__(self) -> str:
        positive_power_strings = (
            dim.replace("K", "\u03B8") + "^" + str(pow)
            for dim, pow in self._d.items()
            if pow > 0
        )
        negative_power_strings = (
            dim.replace("K", "\u03B8") + "^" + str(-pow)
            for dim, pow in self._d.items()
            if pow < 0
        )

        numerator_string = "*".join(sorted(positive_power_strings))
        denominator_string = "*".join(sorted(negative_power_strings))

        if numerator_string == "" and denominator_string == "":
            s = ""
        elif numerator_string != "" and denominator_string == "":
            s = numerator_string
        elif numerator_string == "" and denominator_string == "":
            s = f"1/({denominator_string})"
        else:
            s = f"({numerator_string})/({denominator_string})"

        return s.replace("^1", "")

    def __repr__(self) -> str:
        kwargs_str = ",".join(f"{k}={v}" for k, v in self._d.items())
        return f"Dimension({kwargs_str})"


def _preconvert(func):
    def dec(self, other):
        return func(self, other.convert(self.unit))

    return dec


def _precast(func):
    def dec(self, other):
        try:
            return func(self, other)
        except AttributeError:
            return func(self, Quantity(other))

    return dec


class Quantity(collections.abc.Sequence):
    def __init__(
        self,
        value: Optional[Union[int, float, Iterable[int], Iterable[float]]] = 1.0,
        prefactor: Optional[Union[int, float]] = 1.0,
        L: Optional[int] = 0,
        M: Optional[int] = 0,
        T: Optional[int] = 0,
        I: Optional[int] = 0,
        K: Optional[int] = 0,
        N: Optional[int] = 0,
        J: Optional[int] = 0,
    ) -> None:

        self.value = np.array(value)
        self.prefactor = prefactor
        self.dimension = Dimension(L=L, M=M, T=T, I=I, K=K, N=N, J=J)

    @property
    def magnitude(self):
        return self.value * self.prefactor

    @property
    def unit(self) -> Quantity:
        return Quantity(prefactor=self.prefactor, **self.dimension)

    def to_unit(self) -> Quantity:
        # FIXME: how to preserve value dtype?
        return Quantity(prefactor=self.prefactor * self.value, **self.dimension)

    def convert(self, convert_to: Quantity) -> Quantity:
        if self.dimension != convert_to.dimension:
            raise ValueError("Cannot convert quantities with different dimensions.")
        if self.unit != convert_to.unit:
            # NOTE: can't write this as 'self.value *= ...' without causing numpy unsafe casting error in some cases
            return Quantity(
                value=self.magnitude / convert_to.prefactor,
                prefactor=convert_to.prefactor,
                **convert_to.dimension,
            )
        return Quantity(self.value, self.prefactor, **self.dimension)

    @property
    def T(self):
        return self.value.T * self.unit

    def __getattr__(self, name: str) -> Any:
        if name.startswith("__") and name.endswith("__"):
            raise AttributeError
        # if not hasattr(self,name):
        return getattr(self.value, name)

    __array_priority__ = 1000

    # ADDITION

    @_precast
    @_preconvert
    def __add__(self, other: Union[Quantity, int, float]) -> Quantity:
        return Quantity(self.value + other.value, self.prefactor, **self.dimension)

    @_precast
    @_preconvert
    def __radd__(self, other: Union[Quantity, int, float]) -> Quantity:
        return Quantity(other.value + self.value, self.prefactor, **self.dimension)

    @_precast
    @_preconvert
    def __iadd__(self, other: Union[Quantity, int, float]) -> Quantity:
        self.value += other.value

    # SUBTRACTION

    @_precast
    @_preconvert
    def __sub__(self, other: Union[Quantity, int, float]) -> Quantity:
        return Quantity(self.value - other.value, self.prefactor, **self.dimension)

    @_precast
    @_preconvert
    def __rsub__(self, other: Union[Quantity, int, float]) -> Quantity:
        return Quantity(other.value - self.value, self.prefactor, **self.dimension)

    @_precast
    @_preconvert
    def __isub__(self, other: Union[Quantity, int, float]) -> Quantity:
        self.value -= other.convert(self.unit).value

    # MULTIPLICATION

    @_precast
    def __mul__(self, other: Union[Quantity, float]) -> Quantity:
        return Quantity(
            self.value * other.value,
            self.prefactor * other.prefactor,
            **(self.dimension * other.dimension),
        )

    @_precast
    def __rmul__(self, other: Union[Quantity, int, float]) -> Quantity:
        return Quantity(
            other.value * self.value,
            other.prefactor * self.prefactor,
            **(self.dimension * other.dimension),
        )

    @_precast
    def __imul__(self, other: Union[Quantity, int, float]) -> Quantity:
        self.value *= other.value
        self.prefactor *= other.prefactor
        self.dimension *= other.dimension

        return self

    # MATRIX MULTIPLICATION

    @_precast
    def __matmul__(self, other: Union[Quantity, float]) -> Quantity:
        return Quantity(
            self.value @ other.value,
            self.prefactor * other.prefactor,
            **(self.dimension * other.dimension),
        )

    @_precast
    def __rmatmul__(self, other: Union[Quantity, int, float]) -> Quantity:
        return Quantity(
            other.value @ self.value,
            other.prefactor * self.prefactor,
            **(self.dimension * other.dimension),
        )

    @_precast
    def __imatmul__(self, other: Union[Quantity, int, float]) -> Quantity:
        self.value @= other.value
        self.prefactor *= other.prefactor
        self.dimension *= other.dimension

        return self

    # DIVISION

    @_precast
    def __truediv__(self, other: Union[Quantity, int, float]) -> Quantity:
        return Quantity(
            self.value / other.value,
            self.prefactor / other.prefactor,
            **(self.dimension / other.dimension),
        )

    @_precast
    def __rtruediv__(self, other: Union[Unit, int, float]) -> Quantity:
        return Quantity(
            other.value / self.value,
            other.prefactor / self.prefactor,
            **(other.dimension / self.dimension),
        )

    @_precast
    def __itruediv__(self, other: Union[Unit, int, float]) -> Quantity:
        self.value /= other.value
        self.prefactor /= other.prefactor
        self.dimension /= other.dimension

        return self

    # EXPONENTIATION

    def __pow__(self, other: Union[int, float]) -> Quantity:
        return Quantity(
            self.value ** other, self.prefactor ** other, **(self.dimension ** other)
        )

    def __ipow__(self, other: Union[int, float]) -> Quantity:
        self.value **= other
        self.prefactor **= other
        self.dimension **= other

    # COMPARISON

    @_precast
    def __eq__(
        self, other: Union[Quantity, int, float, Iterable[int], Iterable[float]]
    ) -> bool:
        # np.array_equal works on any array_like and also single values, which should cover all reasonable types for self.value
        return (
            np.array_equal(self.value, other.value)
            and self.prefactor == other.prefactor
            and self.dimension == other.dimension
        )

    @_precast
    @_preconvert
    def __lt__(self, other: Union[Quantity, int, float]) -> bool:
        return self.magnitude < other.magnitude

    @_precast
    @_preconvert
    def __le__(self, other: Union[Quantity, int, float]) -> bool:
        return self.magnitude <= other.magnitude

    @_precast
    @_preconvert
    def __gt__(self, other: Union[Quantity, int, float]) -> bool:
        return self.magnitude > other.magnitude

    @_precast
    @_preconvert
    def __ge__(self, other: Union[Quantity, int, float]) -> bool:
        return self.magnitude >= other.magnitude

    # UNARY

    def __neg__(self) -> Quantity:
        return Quantity(-self.value, self.prefactor, **self.dimension)

    def __pos__(self) -> Quantity:
        return Quantity(+self.value, self.prefactor, **self.dimension)

    def __abs__(self) -> Quantity:
        # FIXME: should abs apply to the prefactor too?
        return Quantity(abs(self.value), self.prefactor, **self.dimension)

    def __invert__(self) -> Quantity:
        # FIXME: should ~ apply to the prefactor too?
        return Quantity(~self.value, self.prefactor, **self.dimension)

    # OTHER

    def __getitem__(self, index):
        return Quantity(self.value[index], self.prefactor, **self.dimension)

    def __len__(self):
        return len(self.value)

    def __hash__(self):
        # FIXME: check this is a good hash method, esp. wrt performance
        return hash((self.value.data.tobytes(), self.dimension))

    def __str__(self) -> str:
        dimension_string = (
            str(self.dimension)
            .replace("L", "m")
            .replace("M", "kg")
            .replace("T", "s")
            .replace("I", "A")
            .replace("\u03B8", "K")
            .replace("N", "mol")
            .replace("J", "cd")
        )

        return f"{self.value*self.prefactor} {dimension_string}"

    def __repr__(self) -> str:
        dim_str = ",".join(f"{k}={v}" for k, v in self.dimension.items())

        return f"Quantity(value={self.value},prefactor={self.prefactor},{dim_str})"



import datetime
from distutils.command.clean import clean
import hashlib
import json
import os
import pathlib
import re
import setuptools
import shutil
import subprocess
from configparser import ConfigParser


class Clean(clean):
    user_options = [("env=", None, "Conda environment to be removed during clean.")]

    def initialize_options(self):
        super().initialize_options()
        self.env = None

    def remove_paths(self, *paths):
        for p in paths:
            try:
                p.unlink()
            except OSError:
                try:
                    shutil.rmtree(p)
                except OSError:
                    continue

    def run(self):
        super().run()
        self.remove_paths(
            pathlib.Path("build"),
            pathlib.Path("conda-build").resolve(),
            pathlib.Path("docs"),
            pathlib.Path(".pytest_cache"),
            pathlib.Path(".coverage"),
            pathlib.Path("coverage.xml"),
            pathlib.Path("dist"),
            *pathlib.Path("src").glob("*.egg-info"),
            *pathlib.Path(".").glob(".coverage*"),
            *pathlib.Path(".").rglob("__pycache__"),
        )
        subprocess.run(["conda", "env", "remove", "-n", self.env])


class Codecov(setuptools.Command):
    user_options = [("codecov-token=", None, "Token for codecov.io coverage upload.")]

    def initialize_options(self):
        self.codecov_token = None

    def finalize_options(self):
        pass

    def run(self):
        subprocess.run(["codecov", "-t", self.codecov_token])


class CondaBuild(setuptools.Command):
    user_options = [
        ("build-dir=", None, "Build directory which will contain built conda package."),
        ("channels=", None, "Conda channels from which to install dependencies"),
        ("recipe=", None, "Directory containing conda recipe."),
    ]

    def initialize_options(self):
        self.build_dir = None
        self.channels = None
        self.recipe = None

    def finalize_options(self):
        pass

    def hash(self):
        metadata = self.distribution.metadata
        # adapted from https://stackoverflow.com/questions/22058048/hashing-a-file-in-python
        m = hashlib.sha256()

        with open(
            pathlib.Path("dist/" + metadata.name + "-" + metadata.version + ".tar.gz"),
            "rb",
        ) as f:
            while True:
                data = f.read(65536)
                if not data:
                    break
                m.update(data)

        return m.hexdigest()

    def run(self):
        channels = " ".join(
            f"-c {c.replace(' ','')}"
            for c in self.channels.splitlines()
            if c.strip() != ""
        ).split()
        subprocess.run(
            ["conda", "build", self.recipe, "--output-folder", self.build_dir]
            + channels,
            env=dict(os.environ, PYPI_TARBALL_SHA_256=self.hash()),
        )


class DeployConda(setuptools.Command):
    user_options = [
        ("build-dir=", None, "Build directory containing built conda package."),
        ("username=", None, "Anaconda username for conda package upload"),
        ("password=", None, "Anaconda password for conda package upload"),
    ]

    def initialize_options(self):
        self.build_dir = None
        self.username = None
        self.password = None

    def finalize_options(self):
        pass

    def run(self):
        package_name = self.distribution.metadata.name
        yes = subprocess.Popen(["yes"], stdout=subprocess.PIPE)
        subprocess.run(
            [
                "anaconda",
                "login",
                "--username",
                self.username,
                "--password",
                self.password,
            ],
            stdin=yes.stdout,
        )

        pwd = pathlib.Path(self.build_dir).absolute()
        installed_package_info = subprocess.check_output(
            [
                "conda",
                "search",
                package_name,
                "--offline",
                "-c",
                "file://" + str(pwd),
                "--json",
            ]
        )
        [d,] = json.loads(installed_package_info)[package_name]
        url = re.match(r"file://(?P<url>.*)", d["url"]).group("url")
        subprocess.run(["anaconda", "upload", url])


class DeployPyPI(setuptools.Command):
    user_options = [
        ("pypi-token=", None, "Token for PyPI package upload"),
    ]

    def initialize_options(self):
        self.pypi_token = None

    def finalize_options(self):
        pass

    def run(self):
        subprocess.run(
            [
                "python",
                "-m",
                "twine",
                "upload",
                *pathlib.Path("dist").rglob("*"),
                "--username",
                "__token__",
                "--password",
                self.pypi_token,
                "--non-interactive",
                "--skip-existing",
            ]
        )


class Docs(setuptools.Command):
    user_options = [
        ("builder=", None, "FIXME"),
        ("extensions=", None, "FIXME"),
        ("params=", None, "FIXME"),
        ("template-dir=", None, "FIXME"),
    ]

    def initialize_options(self):
        self.builder = None
        self.extensions = None
        self.params = None
        self.template_dir = None

    def finalize_options(self):
        pass

    def run(self):
        metadata = self.distribution.metadata
        name = metadata.name
        author = metadata.author
        version = metadata.version

        extensions = ",".join(
            ext.replace(" ", "")
            for ext in self.extensions.splitlines()
            if ext.strip() != ""
        )
        params = " ".join(
            f"-d {ps.replace(' ','')}"
            for ps in self.params.splitlines()
            if ps.strip() != ""
        ).split()

        subprocess.run(
            [
                "sphinx-quickstart",
                "docs",
                "--sep",
                "-p",
                name,
                "-a",
                author,
                "-r",
                version,
                "-l",
                "en",
                "--extensions",
                extensions,
                "-t",
                pathlib.Path(self.template_dir),
                "-d",
                "module_path=" + str(pathlib.Path("src").resolve()),
                "-d",
                "name=" + name,
                "-d",
                "year=" + str(datetime.datetime.now().year),
                "-d",
                "author=" + author,
                "-d",
                "version=" + version,
                "-d",
                "release=" + version,
            ]
            + params
        )

        subprocess.run(
            [
                "sphinx-apidoc",
                pathlib.Path("src/" + name).resolve(),
                "-f",
                "-e",
                "-o",
                pathlib.Path("docs/source"),
                "-t",
                pathlib.Path(self.template_dir),
            ]
        )
        subprocess.run(["make", "-C", "docs", "clean"] + self.builder.split())


class Env(setuptools.Command):
    user_options = [
        ("channels=", None, "Conda channels from which to install dependencies"),
        ("env-name=", None, "Name of conda environment to be created"),
    ]

    def initialize_options(self):
        self.channels = None
        self.env_name = None

    def finalize_options(self):
        pass

    def run(self):
        package_name = self.distribution.metadata.name

        options = setuptools.config.read_configuration("setup.cfg").get("options")
        deps = [
            "python" + str(options.get("python_requires")),
            *options.get("install_requires"),
            *options.get("tests_require"),
            *options.get("extras_require")["dev"],
        ]

        channels = " ".join(
            f"-c {c.replace(' ','')}"
            for c in self.channels.splitlines()
            if c.strip() != ""
        ).split()

        subprocess.run(["conda", "create", "-n", self.env_name, "-y"] + deps + channels)


class InstallLocal(setuptools.Command):
    user_options = [
        (
            "build-dir=",
            None,
            "Build directory containing built conda package. Ignored if --editable is present.",
        ),
        ("channels=", None, "Conda channels from which to install dependencies"),
        ("editable", None, "If true, install editable package using conda develop."),
    ]

    def initialize_options(self):
        self.build_dir = None
        self.channels = None
        self.editable = False

    def finalize_options(self):
        pass

    def run(self):
        if self.editable:
            subprocess.run(["conda", "develop", "src"])
        else:
            package_name = self.distribution.metadata.name
            installed_package_info = subprocess.check_output(
                ["conda", "list", package_name, "--json"]
            )
            if json.loads(installed_package_info) == []:
                pwd = pathlib.Path(self.build_dir).absolute()
                channels = " ".join(
                    f"-c {c.replace(' ','')}"
                    for c in self.channels.splitlines()
                    if c.strip() != ""
                ).split()
                subprocess.run(
                    ["conda", "install", package_name, "-c", "file://" + str(pwd), "-y"]
                    + channels
                )


class Lint(setuptools.Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        subprocess.run(["black", "src", "tests", "setup.py"])


class Tag(setuptools.Command):
    user_options = [
        ("remote=", None, "Name of remote git repository to which tag will be pushed.")
    ]

    def initialize_options(self):
        self.remote = None

    def finalize_options(self):
        pass

    def run(self):
        version = self.distribution.metadata.version
        tag = "v" + version
        subprocess.run(["git", "tag", "-a", tag, "-m", "'" + tag + "'"])
        subprocess.run(["git", "push", self.remote, tag])


class Test(setuptools.Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        subprocess.check_call(["pytest"])


class UpdateVersion(setuptools.Command):
    user_options = [
        ("type=", None, "     Version number update type: major, minor, or patch.")
    ]

    def initialize_options(self):
        self.type = None

    def finalize_options(self):
        pass

    def major(self, m):
        s, *_ = m.groups()
        return f"{int(s)+1}.0.0"

    def minor(self, m):
        s1, s2, *_ = m.groups()
        return f"{s1}.{int(s2)+1}.0"

    def patch(self, m):
        s1, s2, s3 = m.groups()
        return f"{s1}.{s2}.{int(s3)+1}"

    def run(self):
        version = self.distribution.metadata.version
        v = re.sub(r"(\d+)\.(\d+)\.(\d+)", getattr(self, self.type), version)
        cfg_parser = ConfigParser()
        cfg_parser.read("setup.cfg")
        cfg_parser["metadata"]["version"] = v
        with open("setup.cfg", "w") as configfile:
            cfg_parser.write(configfile)


# --------------------------------------- #

setuptools.setup(
    cmdclass={
        "clean": Clean,
        "codecov": Codecov,
        "conda_build": CondaBuild,
        "deploy_conda": DeployConda,
        "deploy_pypi": DeployPyPI,
        "docs": Docs,
        "env": Env,
        "install_local": InstallLocal,
        "lint": Lint,
        "tag": Tag,
        "test": Test,
        "version": UpdateVersion,
    }
)



from __future__ import annotations
from typing import Dict, Iterable, Optional, Tuple, Union

import ast
import materia as mtr
import pathlib
import re
import shlex
import subprocess

__all__ = [
    "Engine",
    "CCDC",
    "FragIt",
    "GPAW",
    "Multiwfn",
    "Openbabel",
    "Packmol",
    "QChem",
    "XTB",
]


class Engine:
    def __init__(
        self,
        executable: str,
        num_processors: Optional[int] = None,
        num_threads: Optional[int] = None,
        arguments: Optional[Iterable[str]] = None,
    ) -> None:

        self.executable = executable
        self.num_processors = num_processors
        self.num_threads = num_threads
        self.arguments = arguments or []

    def env(self) -> Dict[str, str]:
        return None

    def command(
        self,
        inp: str,
        out: str,
        work_dir: str,
        arguments: Optional[Iterable[str]] = None,
    ) -> str:
        args = list(self.arguments) + list(arguments or [])
        arg_str = " ".join(args)
        # FIXME: shlex.quote should be used but it doesn't work...
        return shlex.split(f"{self.executable} {arg_str}")

    def execute(self, io: materia.IO, arguments: Optional[Iterable[str]] = None) -> str:
        with io() as _io:
            cmd = self.command(_io.inp, _io.out, _io.work_dir, arguments)
            with open(_io.inp, "r") as inp:
                with open(_io.out, "w") as out:
                    env = self.env()
                    if env is None:
                        subprocess.call(cmd, stdin=inp, stdout=out, cwd=io.work_dir)
                    else:
                        subprocess.call(
                            cmd, stdin=inp, stdout=out, env=self.env(), cwd=io.work_dir
                        )

            with open(_io.out, "r") as f:
                return "".join(f.readlines())


class CCDC(Engine):
    def __init__(
        self,
        ccdc_root: str,
        num_processors: Optional[int] = None,
        num_threads: Optional[int] = None,
        arguments: Optional[Iterable[str]] = None,
    ) -> None:
        self.ccdc_root = mtr.expand(ccdc_root)
        # FIXME: generalize past 2019 version of CCDC code
        executable = mtr.expand(
            pathlib.Path(
                self.ccdc_root, "Python_API_2019", "miniconda", "bin", "python"
            )
        )
        super().__init__(executable, num_processors, num_threads, arguments)

    def env(self) -> Dict[str, str]:
        # FIXME: generalize past 2019 version of CCDC code
        return {"CSDHOME": mtr.expand(pathlib.Path(self.ccdc_root, "CSD_2019"))}


class FragIt(Engine):
    def __init__(
        self,
        executable: Optional[str] = "fragit",
        arguments: Optional[Iterable[str]] = None,
    ) -> None:
        super().__init__(executable=executable, arguments=arguments)

    def command(
        self,
        inp: str,
        out: str,
        work_dir: str,
        arguments: Optional[Iterable[str]] = None,
    ) -> str:
        args = list(self.arguments) + list(arguments or [])
        arg_str = " ".join(args)
        # FIXME: shlex.quote should be used but it doesn't work...
        return shlex.split(f"{self.executable} {inp} {arg_str}")

    def execute(self, io_params: materia.IO) -> str:
        with io_params() as io:
            cmd = self.command(io.inp, io.out, io.work_dir)
            with open(io.inp, "r") as inp:
                with open(io.out, "w") as out:
                    subprocess.call(
                        cmd, stdin=inp, stdout=out, env=self.env(), cwd=io.work_dir
                    )

            with open(io.out, "r") as f:
                return "".join(f.readlines())


class GPAW(Engine):
    def __init__(
        self,
        executable: Optional[str] = "gpaw",
        num_processors: Optional[int] = None,
        num_threads: Optional[int] = None,
        arguments: Optional[Iterable[str]] = None,
    ) -> None:
        super().__init__(executable, num_processors, num_threads, arguments)

    def execute(self, structure_file):
        raise NotImplementedError


class Multiwfn(Engine):
    def __init__(
        self,
        executable: Optional[str] = "Multiwfn",
        num_processors: Optional[int] = None,
        num_threads: Optional[int] = None,
        arguments: Optional[Iterable[str]] = None,
    ) -> None:
        super().__init__(executable, num_processors, num_threads, arguments)

    def execute(self, io_params: materia.IO) -> str:
        env = self.env()

        with io_params() as io:
            cmd = self.command(io.inp, io.out, io.work_dir)

            with open(io.inp, "r") as inp:
                input_lines = "".join(inp.readlines())

            if self.num_processors is not None:
                input_lines = ["1000\n", "10\n", f"{self.num_processors}\n"]

            with open(io.out, "w") as out:
                # For Multiwfn 3.6
                # FIXME: printf is system-specific - is there any other way to pipe in input_lines?
                pipe_command_string = subprocess.Popen(
                    args=["printf", input_lines],
                    stdout=subprocess.PIPE,
                    encoding="utf-8",
                    env=env,
                )
                multiwfn = subprocess.Popen(
                    cmd,
                    stdin=pipe_command_string.stdout,
                    stdout=out,
                    stderr=subprocess.STDOUT,
                    encoding="utf-8",
                    env=env,
                )
                pipe_command_string.stdout.close()
                multiwfn.communicate()


class Openbabel(Engine):
    def __init__(
        self,
        executable: Optional[str] = "obabel",
        arguments: Optional[Iterable[str]] = None,
    ) -> None:
        super().__init__(executable=executable, arguments=arguments)

    def command(
        self,
        inp: str,
        out: str,
        work_dir: str,
        arguments: Optional[Iterable[str]] = None,
    ) -> str:
        args = list(self.arguments) + list(arguments or [])
        arg_str = " ".join(args)
        # FIXME: shlex.quote should be used but it doesn't work...
        return shlex.split(f"{self.executable} {inp} {arg_str}")


class Packmol(Engine):
    def __init__(self, executable: Optional[str] = "packmol") -> None:
        super().__init__(executable=executable)


class QChem(Engine):
    def __init__(
        self,
        scratch_dir: Optional[str] = None,
        qcenv: Optional[str] = None,
        executable: Optional[str] = "qchem",
        num_processors: Optional[int] = None,
        num_threads: Optional[int] = None,
        arguments: Optional[Iterable[str]] = None,
    ) -> None:
        self.scratch_dir = mtr.expand(scratch_dir) if scratch_dir is not None else None
        self.qcenv = shlex.quote(mtr.expand(qcenv)) if qcenv is not None else None
        super().__init__(
            executable=executable,
            num_processors=num_processors,
            num_threads=num_threads,
            arguments=arguments,
        )

    def env(self) -> Dict[str, str]:
        if self.scratch_dir is None and self.qcenv is None:
            return None

        if self.qcenv is not None:
            # FIXME: shell=True needs to be avoided!!
            d = ast.literal_eval(
                re.match(
                    r"environ\((.*)\)",
                    subprocess.check_output(
                        f". {self.qcenv}; python -c 'import os; print(os.environ)'",
                        shell=True,
                    )
                    .decode()
                    .strip(),
                ).group(1)
            )
        else:
            d = {}

        if self.scratch_dir is not None:
            d["QCSCRATCH"] = mtr.expand(self.scratch_dir)

        return d

    def command(
        self,
        inp: str,
        out: str,
        work_dir: str,
        arguments: Optional[Iterable[str]] = None,
    ) -> str:
        cmd = [self.executable]
        if self.arguments is not None:
            cmd.extend(list(self.arguments) + list(arguments or []))
        if self.num_processors is not None:
            cmd.append(f"-np {self.num_processors}")
        if self.num_threads is not None:
            cmd.append(f"-nt {self.num_threads}")

        # FIXME: shlex.quote should be used but it doesn't work...
        return shlex.split(" ".join(cmd + [inp, out]))


class XTB(Engine):
    def __init__(
        self,
        executable: Optional[str] = "xtb",
        arguments: Optional[Iterable[str]] = None,
    ) -> None:
        super().__init__(executable=executable, arguments=arguments)

    def command(
        self,
        out: str,
        work_dir: str,
        coord: str,
        arguments: Optional[Iterable[str]] = None,
    ) -> str:
        args = list(self.arguments) + list(arguments or [])
        arg_str = " ".join(args)
        # FIXME: shlex.quote should be used but it doesn't work...
        return shlex.split(f"{self.executable} {coord} {arg_str}")

    def execute(
        self, coord: str, io: materia.IO, arguments: Optional[Iterable[str]] = None
    ) -> str:
        with io() as _io:
            cmd = self.command(_io.out, _io.work_dir, mtr.expand(coord), arguments)
            with open(_io.out, "w") as out:
                env = self.env()
                if env is None:
                    subprocess.call(
                        cmd, stdout=out, stderr=subprocess.STDOUT, cwd=io.work_dir
                    )
                else:
                    subprocess.call(
                        cmd,
                        stdout=out,
                        stderr=subprocess.STDOUT,
                        env=self.env(),
                        cwd=io.work_dir,
                    )

            with open(_io.out, "r") as f:
                return "".join(f.readlines())



from .actions import *
from .dataseries import *
from .engines import *
from .handler import *
from .io import *
from .molecule import *
from .properties import *
from .quantity import *
from .symmetry import *
from .tasks import *
from .units import *
from .utils import *
from .workflow import *



import itertools
import numpy as np
import scipy.linalg
import math

import materia

# SEE https://iopscience.iop.org/book/978-1-6817-4637-1/chapter/bk978-1-6817-4637-1ch1, https://math.stackexchange.com/questions/693414/reflection-across-the-plane, https://github.com/psi4/psi4/blob/master/psi4/src/psi4/libmints/molecule.cc, http://www.pci.tu-bs.de/aggericke/PC4e/Kap_IV/Matrix_Symm_Op.htm


class SymmetryAnalyzer:
    def symmetry_operations(self, molecule, symprec=5, graph=False):
        if graph:
            permutations = {
                hash(P.data.tobytes()): P
                for P in self._automorphism_iterator(molecule=molecule, symprec=symprec)
            }

            for P in permutations.values():
                permutations[hash(P.T.data.tobytes())] = P.T
            for P1, P2 in itertools.combinations(permutations.values(), r=2):
                P3 = P1 @ P2
                permutations[hash(P3.data.tobytes())] = P3
                permutations[hash(P3.T.data.tobytes())] = P3.T

            X = molecule.structure.centered_atomic_positions.value
            pinv_X = np.linalg.pinv(X)

            # FIXME: need to check that these rotations are actually symmetries before returning them!
            return tuple(
                SymmetryOp(matrix=scipy.linalg.polar(X @ P @ pinv_X)[0])
                for P in permutations.values()
            )
        else:
            I = molecule.structure.inertia_tensor / np.trace(
                molecule.structure.inertia_tensor
            )
            principal_moments, principal_axes = scipy.linalg.eig(I)
            (m1, m2, m3) = principal_moments.round(symprec)

            symmetry_operations = []

            for p1, p2 in itertools.combinations(
                molecule.structure.centered_atomic_positions.value.T, r=2
            ):
                axis = p1 - p2
                axis /= scipy.linalg.norm(axis)

                R = Reflection(axis=axis)

                if R.is_symmetry_of(molecule, symprec=symprec):
                    symmetry_operations.append(R)
                print(len(symmetry_operations))
                axis = p1 + p2
                axis_norm = scipy.linalg.norm(axis)
                if not np.isclose(axis_norm, 0):
                    axis /= axis_norm
                    R2 = ProperRotation(order=2, axis=axis)
                    R3 = ProperRotation(order=2, axis=axis)
                    R4 = ProperRotation(order=2, axis=axis)
                    R5 = ProperRotation(order=2, axis=axis)

                    if R2.is_symmetry_of(molecule, symprec=symprec):
                        symmetry_operations.append(R2)
                    if R3.is_symmetry_of(molecule, symprec=symprec):
                        symmetry_operations.append(R3)
                    if R4.is_symmetry_of(molecule, symprec=symprec):
                        symmetry_operations.append(R4)
                    if R5.is_symmetry_of(molecule, symprec=symprec):
                        symmetry_operations.append(R5)

                print(len(symmetry_operations))
            print(len({hash(R.matrix.data.tobytes()): R for R in symmetry_operations}))
            if np.isclose(np.prod((m1, m2, m3)), 0):  # one of the eigenvalues is zero
                return self._analyze_linear_molecule(molecule=molecule)
            elif np.allclose((m1 - m2, m2 - m3), 0):  # all eigenvalues are the same
                return self._analyze_spherical_top(molecule=molecule)
            elif (
                np.sum(np.isclose((m1 - m2, m2 - m3, m1 - m3), 0)) == 0
            ):  # all eigenvalues are different
                # possible groups: C1, Ci, Cs, C2, C2v, C2h, D2, D2h
                for axis in principal_axes.T:
                    R = ProperRotation(order=2, axis=axis)
                    if R.is_symmetry_of(molecule=molecule, symprec=symprec):
                        symmetry_operations.append(R)
                if len(symmetry_operations) == 0:
                    # possible groups: C1, Ci, Cs
                    inv = Inversion()
                    if inv.is_symmetry_of(molecule=molecule, symprec=symprec):
                        # Ci
                        symmetry_operations.append(inv)
                    # elif Reflection(axis=)# check for mirror plane
                elif len(symmetry_operations) == 1:
                    # possible groups: C2, C2v, C2h
                    pass
                else:
                    # possible groups: D2, D2h
                    pass
                print(symmetry_operations)
                return self._analyze_asymmetrical_top(molecule=molecule)
            else:  # two eigenvalues are the same, one is different
                return self._analyze_symmetrical_top(molecule=molecule)

    def _analyze_linear_molecule(self, molecule):
        inv = Inversion()
        if inv.is_symmetry_of(molecule):
            pass  # FIXME: what to return for infinite point group Dinfh?
        else:
            pass  # FIXME: what to return for infinite point group Cinfv?

    # def asymmetrical_top(self, molecule):
    #     for

    def _automorphism_iterator(self, molecule, symprec):
        import networkx as nx

        num_atoms = len(molecule.structure.atoms)

        g = nx.Graph()

        nodes = tuple(
            range(num_atoms + 1)
        )  # node with label num_atoms represents COM node
        edges = tuple(
            (num_atoms, i) for i in nodes[:-1]
        )  # make edge between COM node and every non-COM node

        atomic_symbols = (
            *molecule.structure.atomic_symbols,
            "COM",
        )  # 'COM' is the "atomic symbol" for the COM node
        com_distances = (
            np.linalg.norm(p).round(symprec)
            for p in molecule.structure.centered_atomic_positions.value.T
        )

        g.add_edges_from(edges)
        nx.set_node_attributes(
            G=g, values={n: {"Z": v} for n, v in zip(nodes, atomic_symbols)}
        )
        nx.set_edge_attributes(
            G=g, values={e: {"dist": v} for e, v in zip(edges, com_distances)}
        )

        permutation_iterator = nx.algorithms.isomorphism.GraphMatcher(
            G1=g,
            G2=g,
            node_match=lambda n1, n2: n1["Z"] == n2["Z"],
            edge_match=lambda e1, e2: e1["dist"] == e2["dist"],
        ).isomorphisms_iter()

        yield from (
            self._permutation_matrix(permutation_dict=perm)
            for perm in permutation_iterator
        )

    def _permutation_matrix(self, permutation_dict):
        n = len(permutation_dict) - 1  # number of atoms
        permutation_dict.pop(n)  # remove the COM

        I = np.eye(n, dtype=int)
        return np.hstack(
            [
                I[:, col_ind].reshape(n, -1)
                for _, col_ind in sorted(permutation_dict.items())
            ]
        )


class PointGroup:
    def __init__(self, reps):
        self.reps = tuple(np.atleast_1d(r) for r in reps)

    def cayley_table(self, reps):
        labels = {k: i for i, k in enumerate(reps)}
        num_reps = len(labels)
        ctable = np.zeros(shape=(num_reps, num_reps), dtype=int)

        for P, Q in itertools.combinations(reps.values(), r=2):
            c = np.atleast_1d(P @ Q)
            i = labels[hash(P.data.tobytes())]
            j = labels[hash(Q.data.tobytes())]
            ctable[i, j] = ctable[j, i] = labels[hash(c.data.tobytes())]

        rowset = set(tuple(r) for r in ctable)
        colset = set(tuple(c) for c in ctable.T)

        return ctable


class SymmetryOp:
    def __init__(self, matrix=None, determinant=None, trace=None, axis=None):
        if matrix is not None:
            self.matrix, _ = scipy.linalg.polar(matrix)
        else:
            if axis is None:
                self.matrix, _ = scipy.linalg.polar(
                    determinant * np.eye(3).astype("float64")
                )
            else:
                cos_theta = max(min((trace - determinant) / 2, 1), -1)
                sin_theta = max(min(np.sqrt(1 - cos_theta ** 2), 1), -1)

                u1, u2, u3 = axis
                K = np.array([[0, -u3, u2], [u3, 0, -u1], [-u2, u1, 0]])

                rotation = (
                    np.eye(3) + sin_theta * K + (1 - cos_theta) * (K @ K)
                ).astype("float64")
                reflection = np.eye(3) - (determinant == -1) * 2 * np.outer(axis, axis)

                self.matrix, _ = scipy.linalg.polar(rotation @ reflection)

    @property
    def det(self):
        return int(round(np.linalg.det(self.matrix)))

    @property
    def tr(self):
        return np.trace(self.matrix)

    @property
    def cos_theta(self):
        return max(min((self.tr - np.sign(self.det)) / 2, 1), -1)

    @property
    def axis(self):
        if np.isclose(abs(self.tr), 3):
            return None

        S = self.matrix + self.matrix.T - (self.tr - np.sign(self.det)) * np.eye(3)
        S[np.isclose(S, 0)] = 0
        axis = np.sqrt(
            np.abs(np.diag(S) / (3 - np.sign(self.det) * self.tr))
        )  # scipy.linalg.null_space(R-det*np.eye(3)).squeeze()
        u = self._perpendicular_vector(p=axis)
        axis *= np.sign(np.dot(axis, np.cross(u, self.matrix @ u)))

        return axis

    def _perpendicular_vector(self, p):
        m = np.zeros(p.shape)

        i = (np.ravel(p) != 0).argmax()  # index of the first nonzero element of p
        j = next(
            ind for ind in range(len(np.ravel(p))) if ind != i
        )  # first index of p which is not i
        i, j = (
            np.unravel_index(i, p.shape),
            np.unravel_index(j, p.shape),
        )  # unravel indices for 1x3 arrays m and p

        # make m = np.array([[-py,px,0]]) so np.dot(m,p) = -px*py + px*py = 0
        m[j] = p[i]
        m[i] = -p[j]
        m /= scipy.linalg.norm(m)

        return np.array(m)

    def is_symmetry_of(self, structure, symprec):
        transformed_atomic_positions = (
            self.matrix @ structure.centered_atomic_positions.value
        )
        return set(
            tuple(row) for row in transformed_atomic_positions.T.round(symprec)
        ) == set(
            tuple(row)
            for row in structure.centered_atomic_positions.value.T.round(symprec)
        )


class Inversion(SymmetryOp):
    def __init__(self):
        determinant = -1
        trace = -3
        axis = None
        super().__init__(determinant=determinant, trace=trace, axis=axis)


class Reflection(SymmetryOp):
    def __init__(self, axis):
        determinant = -1
        trace = 1
        super().__init__(determinant=determinant, trace=trace, axis=axis)


class ProperRotation(SymmetryOp):
    def __init__(self, order, axis):
        determinant = 1
        trace = 2 * np.cos(2 * np.pi / order) + determinant
        super().__init__(determinant=determinant, trace=trace, axis=axis)


class ImproperRotation(SymmetryOp):
    def __init__(self, order, axis):
        determinant = -1
        trace = 2 * np.cos(2 * np.pi / order) + determinant
        super().__init__(determinant=determinant, trace=trace, axis=axis)


# class SigmaX(SymmetryOp):
#     def __init__(self):
#         super().__init__(matrix=np.array([[-1,0,0],[0,1,0],[0,0,1]]))
#
# class SigmaY(SymmetryOp):
#     def __init__(self):
#         super().__init__(matrix=np.array([[1,0,0],[0,-1,0],[0,0,1]]))
#
# class SigmaZ(SymmetryOp):
#     def __init__(self):
#         super().__init__(matrix=np.array([[1,0,0],[0,1,0],[0,0,-1]]))
#
# class Inversion(SymmetryOp):
#     def __init__(self):
#         super().__init__(matrix=np.array([[-1,0,0],[0,-1,0],[0,0,-1]]))
#         self.periodicity = 2
#
# class Sigma(SymmetryOp):
#     def __init__(self, axis):
#         matrix = np.eye(3) - 2*np.outer(axis,axis)
#         super().__init__(matrix=matrix)
#         self.periodicity = 2
#
# class CnU(SymmetryOp):
#     def __init__(self, n, axis):
#         u1,u2,u3 = np.asarray(axis).reshape(3,)
#
#         K = np.array([[0,-u3,u2],[u3,0,-u1],[-u2,u1,0]])
#
#         angle = 2*np.pi/n
#
#         matrix = (np.eye(3) + np.sin(angle)*K + (1-np.cos(angle))*(K @ K)).astype('float64')
#
#         super().__init__(matrix=matrix)
#         self.periodicity = n
#
# class C2X(SymmetryOp):
#     def __init__(self):
#         self.matrix = np.array([[1,0,0],[0,np.cos(np.pi),np.sin(np.pi)],[0,-np.sin(np.pi),np.cos(np.pi)]])
#
# class C2Y(SymmetryOp):
#     def __init__(self):
#         self.matrix = np.array([[np.cos(np.pi),0,np.sin(np.pi)],[0,1,0],[-np.sin(np.pi),0,np.cos(np.pi)]])
#
# class C2Z(SymmetryOp):
#     def __init__(self):
#         self.matrix = np.array([[np.cos(np.pi),np.sin(np.pi),0],[-np.sin(np.pi),np.cos(np.pi),0],[0,0,1]])
#
# class C3X(SymmetryOp):
#     def __init__(self):
#         self.matrix = np.array([[1,0,0],[0,np.cos(2*np.pi/3),np.sin(2*np.pi/3)],[0,-np.sin(2*np.pi/3),np.cos(2*np.pi/3)]])
#
# class C3Y(SymmetryOp):
#     def __init__(self):
#         self.matrix = np.array([[np.cos(2*np.pi/3),0,np.sin(2*np.pi/3)],[0,1,0],[-np.sin(2*np.pi/3),0,np.cos(2*2*np.pi/3/3)]])
#
# class C3Z(SymmetryOp):
#     def __init__(self):
#         self.matrix = np.array([[np.cos(2*np.pi/3),np.sin(2*np.pi/3),0],[-np.sin(2*np.pi/3),np.cos(2*np.pi/3),0],[0,0,1]])
#
# class C4X(SymmetryOp):
#     def __init__(self):
#         self.matrix = np.array([[1,0,0],[0,np.cos(np.pi/2),np.sin(np.pi/2)],[0,-np.sin(np.pi/2),np.cos(np.pi/2)]])
#
# class C4Y(SymmetryOp):
#     def __init__(self):
#         self.matrix = np.array([[np.cos(np.pi/2),0,np.sin(np.pi/2)],[0,1,0],[-np.sin(np.pi/2),0,np.cos(np.pi/2)]])
#
# class C4Z(SymmetryOp):
#     def __init__(self):
#         self.matrix = np.array([[np.cos(np.pi/2),np.sin(np.pi/2),0],[-np.sin(np.pi/2),np.cos(np.pi/2),0],[0,0,1]])
#
# class C5X(SymmetryOp):
#     def __init__(self):
#         self.matrix = np.array([[1,0,0],[0,np.cos(2*np.pi/5),np.sin(2*np.pi/5)],[0,-np.sin(2*np.pi/5),np.cos(2*np.pi/5)]])
#
# class C5Y(SymmetryOp):
#     def __init__(self):
#         self.matrix = np.array([[np.cos(2*np.pi/5),0,np.sin(2*np.pi/5)],[0,1,0],[-np.sin(2*np.pi/5),0,np.cos(2*np.pi/5)]])
#
# class C5Z(SymmetryOp):
#     def __init__(self):
#         self.matrix = np.array([[np.cos(2*np.pi/5),np.sin(2*np.pi/5),0],[-np.sin(2*np.pi/5),np.cos(2*np.pi/5),0],[0,0,1]])
#
# class C6X(SymmetryOp):
#     def __init__(self):
#         self.matrix = np.array([[1,0,0],[0,np.cos(np.pi/3),np.sin(np.pi/3)],[0,-np.sin(np.pi/3),np.cos(np.pi/3)]])
#
# class C6Y(SymmetryOp):
#     def __init__(self):
#         self.matrix = np.array([[np.cos(np.pi/3),0,np.sin(np.pi/3)],[0,1,0],[-np.sin(np.pi/3),0,np.cos(np.pi/3)]])
#
# class C6Z(SymmetryOp):
#     def __init__(self):
#         self.matrix = np.array([[np.cos(np.pi/3),np.sin(np.pi/3),0],[-np.sin(np.pi/3),np.cos(np.pi/3),0],[0,0,1]])


def _equivalent_atoms(molecule):
    mol_dict = {
        tuple(p.squeeze()): symb
        for symb, p in zip(
            molecule.structure.atomic_symbols,
            molecule.structure.centered_atomic_positions.value.T,
        )
    }
    elements = tuple(set(mol_dict.values()))

    yield from (
        [position for position, element in mol_dict.items() if element == el]
        for el in elements
    )


def _candidate_symmetry_axes(molecule):
    centroid_lines = (
        np.asarray(position1) + np.asarray(position2)
        for position1, position2 in itertools.chain(
            *(
                itertools.combinations(isoelemental_group, r=2)
                for isoelemental_group in _equivalent_atoms(molecule=molecule)
            )
        )
    )

    return tuple(
        tuple(l / np.linalg.norm(l))
        for l in itertools.chain(
            molecule.structure.centered_atomic_positions.value.T, centroid_lines
        )
    )


def find_symmetry(mol):
    # FIXME: first align principal axes, center on COM
    possible_lines_of_symmetry = _candidate_symmetry_axes(molecule=mol)
    if mol.structure.is_linear:
        if Inversion().is_symmetry_of(mol=mol):
            print("D-infh")
        else:
            print("C-infv")
    # print([CnU(n=6,axis=line).is_symmetry_of(mol=mol) for line in mol.structure.atomic_lines])
    # print([CnU(n=5,axis=line).is_symmetry_of(mol=mol) for line in mol.structure.atomic_lines])
    # print([CnU(n=4,axis=line).is_symmetry_of(mol=mol) for line in mol.structure.atomic_lines])
    # print([CnU(n=3,axis=line).is_symmetry_of(mol=mol) for line in mol.structure.atomic_lines])
    # print([CnU(n=2,axis=line).is_symmetry_of(mol=mol) for line in mol.structure.atomic_lines])
    if (
        sum(
            CnU(n=5, axis=line).is_symmetry_of(molecule=mol)
            for line in possible_lines_of_symmetry
        )
        >= 2
    ):
        if mol.has_symmetry(symmetry_op=Inversion()):
            print("Ih")
        else:
            print("I")
    if (
        sum(
            CnU(n=4, axis=line).is_symmetry_of(molecule=mol)
            for line in possible_lines_of_symmetry
        )
        >= 2
    ):
        if mol.has_symmetry(symmetry_op=Inversion()):
            print("Oh")
        else:
            print("O")
    if (
        sum(
            CnU(n=3, axis=line).is_symmetry_of(molecule=mol)
            for line in possible_lines_of_symmetry
        )
        == 4
    ):
        if (
            sum(
                mol.has_symmetry(symmetry_op=Sigma(axis=plane_normal))
                for plane_normal in mol.atom_planes
            )
            >= 1
        ):
            if mol.has_symmetry(symmetry_op=Inversion()):
                print("Th")
            else:
                print("Td")
        else:
            print("T")
    print(
        [
            CnU(n=2, axis=line).is_symmetry_of(molecule=mol)
            for line in possible_lines_of_symmetry
        ]
    )
    print(possible_lines_of_symmetry)
    has_cs = [
        sum(
            CnU(n=n, axis=line).is_symmetry_of(molecule=mol)
            for line in possible_lines_of_symmetry
        )
        for n in (2, 3, 4, 5, 6)
    ]
    if sum(has_cs) == 0:
        if (
            sum(
                mol.has_symmetry(symmetry_op=Sigma(axis=plane_normal))
                for plane_normal in mol.atom_planes
            )
            >= 1
        ):
            print("Cs")
        else:
            if mol.has_symmetry(symmetry_op=Inversion()):
                print("Ci")
            else:
                print("C1")
    else:
        max_n = 6 - np.argmax(has_cs[::-1])
        if max_n <= 1:
            pass
